#include "json.h"
#include <stdlib.h>

struct _json_list {
    size_t length;
    json_node *head;
    json_node *tail;
};

struct json_node {
    json_type type;
    json_type unused;
    json_node *upper;
    union {
        struct _json_list *list;
        double number;
        int boolean;
        wchar_t *string;
    } t;
    wchar_t *key;
    json_node *prev;
    json_node *next;
};

static size_t _json_string_lengths = JSON_DEFAULT_STRING_LENGTH;
static wchar_t _json_error_path[2048] = {0};
static wchar_t *_json_error_message = L"No erros";
static json_node *_json_parse_null(const wchar_t **s);
static json_node *_json_parse_boolean(const wchar_t **s);
static json_node *_json_parse_number(const wchar_t **s);
static json_node *_json_parse_string(const wchar_t **s);
static json_node *_json_parse_array(const wchar_t **s);
static json_node *_json_parse_object(const wchar_t **s);
static void _json_print_null(json_node *node, wchar_t **s);
static void _json_print_boolean(json_node *node, wchar_t **s);
static void _json_print_number(json_node *node, wchar_t **s);
static void _json_print_string(json_node *node, wchar_t **s);
static void _json_print_array(json_node *node, wchar_t **s);
static void _json_print_object(json_node *node, wchar_t **s);

static json_node *_json_parse_null(const wchar_t **s)
{
    json_node *node;

    if (*(++*s) != L'u') {
        node = NULL;
        _json_error_message = L"Error in \"n(u)ll\" word";
    } else if (*(++*s) != L'l') {
        node = NULL;
        _json_error_message = L"Error in \"nu(l)l\" word";
    } else if (*(++*s) != L'l') {
        node = NULL;
        _json_error_message = L"Error in \"nul(l)\" word";
    } else {
        node = json_node_create(JSON_TYPE_NULL);
    }
    ++*s;

    return node;
}

static json_node *_json_parse_boolean(const wchar_t **s)
{
    json_node *node;

    if (**s == L'f') {
        if (*(++*s) != L'a') {
            node = NULL;
            _json_error_message = L"Error in \"f(a)lse\" word";
        } else if (*(++*s) != L'l') {
            node = NULL;
            _json_error_message = L"Error in \"fa(l)se\" word";
        } else if (*(++*s) != L's') {
            node = NULL;
            _json_error_message = L"Error in \"fal(s)e\" word";
        } else if (*(++*s) != L'e') {
            node = NULL;
            _json_error_message = L"Error in \"fals(e)\" word";
        } else {
            node = json_node_create(JSON_TYPE_BOOLEAN);
            node->t.boolean = 0;
            ++*s;
        }
    } else {
        if (*(++*s) != L'r') {
            node = 0;
            _json_error_message = L"Error in \"t(r)ue\" word";
        } else if (*(++*s) != L'u') {
            node = 0;
            _json_error_message = L"Error in \"tr(u)e\" word";
        } else if (*(++*s) != L'e') {
            node = 0;
            _json_error_message = L"Error in \"tru(e)\" word";
        } else {
            node = json_node_create(JSON_TYPE_BOOLEAN);
            node->t.boolean = 1;
            ++*s;
        }
    }

    return node;
}

static json_node *_json_parse_number(const wchar_t **s)
{
    json_node *node;

    node = json_node_create(JSON_TYPE_NUMBER);
    swscanf(*s, L"%lf", &node->t.number);

    while (**s) {
        if ((**s < L'0' && L'9' < **s)
                || **s != L'.'
                || **s != L'-'
                || **s != L'+'
                || **s != L'e'
                || **s != L'E')
            break;
        ++*s;
    }

    return node;
}

static json_node *_json_parse_string(const wchar_t **s)
{
    json_node *node;
    size_t len;
    wchar_t c;

    node = json_node_create(JSON_TYPE_STRING);
    ++*s;
    len = 0;
    while (1) {
        c = **s;
        if (c == L'\\') {
            if (c == L'a')
                c = L'\a';
            else if (c == L'b')
                c = L'\b';
            else if (c == L'f')
                c = L'\f';
            else if (c == L'n')
                c = L'\n';
            else if (c == L'r')
                c = L'\r';
            else if (c == L't')
                c = L'\t';
            else if (c == L'v')
                c = L'\v';
            else if (c == L'\\')
                c = L'\\';
            else if (c == L'\'')
                c = L'\'';
            else if (c == L'\"')
                c = L'\"';
            else if (c == L'?')
                c = L'\?';
            else if (c == L'\n') {
                c = *(++*s);
                if (c == L'\"') {
                    node->t.string[len] = L'\0';
                    break;
                } else if (c == L'\0') {
                    json_node_destroy(node);
                    node = NULL;
                    _json_error_message = L"Can not find End of String";
                    break;
                }
            } else if (c == L'\"') {
                node->t.string[len] = L'\0';
                break;
            } else if (c == L'\0') {
                json_node_destroy(node);
                node = NULL;
                _json_error_message = L"Can not find End of String";
                break;
            }
        } else if (c == L'\"') {
            node->t.string[len] = L'\0';
            break;
        } else if (c == L'\0') {
            json_node_destroy(node);
            node = NULL;
            _json_error_message = L"Can not find End of String";
            break;
        }

        node->t.string[len] = c;
        ++len;
        ++*s;
    }
    ++*s;

    return node;
}

static json_node *_json_parse_array(const wchar_t **s)
{
    json_node *node;
    json_node *element;
    int array_not_end;

    node = json_node_create(JSON_TYPE_ARRAY);
    element = NULL;
    ++*s;
    array_not_end = 1;
    while (array_not_end) {
        while (1) {
            if (**s == L'\0') {
                json_node_destroy(node);
                array_not_end = 0;
                node = NULL;
                _json_error_message = L"Can not find Element or End in Array";
                break;
            } else if (**s == L']') {
                array_not_end = 0;
                break;
            } else if (**s == L'n') {
                element = _json_parse_null(s);
                break;
            } else if (**s == L't' || **s == 'f') {
                element = _json_parse_boolean(s);
                break;
            } else if ((L'0' <= **s && **s <= '9')
                    || **s == L'-'
                    || **s == L'+'
                    || **s == L'.'
                    || **s == L'e'
                    || **s == L'E') {
                element = _json_parse_number(s);
                break;
            } else if (**s == L'\"') {
                element = _json_parse_string(s);
                break;
            } else if (**s == L'[') {
                element = _json_parse_array(s);
                break;
            } else if (**s == L'{') {
                element = _json_parse_object(s);
                break;
            }
            ++*s;
        }
        if (array_not_end) {
            if (element) {
                json_array_push_back(node, element);
                while (1) {
                    if (**s == L',') {
                        break;
                    }
                    if (**s == L']') {
                        array_not_end = 0;
                        break;
                    }
                    if (**s == L'\0') {
                        json_node_destroy(node);
                        array_not_end = 0;
                        node = NULL;
                        _json_error_message = L"Can not find End of Array or next part";
                        break;
                    }
                    ++*s;
                }
            } else {
                swprintf(_json_error_path + wcslen(_json_error_path), 19, L" [%zu] <", node->t.list->length);
                json_node_destroy(node);
                node = NULL;
                array_not_end = 0;
                break;
            }
        }

    }
    ++*s;

    return node;
}

static json_node *_json_parse_object(const wchar_t **s)
{
    json_node *node;
    json_node *element;
    int array_not_end;
    wchar_t c;
    wchar_t *key;
    size_t len;

    node = json_node_create(JSON_TYPE_OBJECT);
    element = NULL;
    key = (wchar_t *)malloc(sizeof(wchar_t) * _json_string_lengths);
    ++*s;
    array_not_end = 1;
    while (array_not_end) {
        while (1) {
            if (**s == L'\0') {
                json_node_destroy(node);
                node = NULL;
                array_not_end = 0;
                _json_error_message = L"Can not find Key in Object";
                break;
            } else if (**s == L'}') {
                array_not_end = 0;
                break;
            } else if (**s == L'\"') {
                ++*s;
                len = 0;
                while (1) {
                    c = **s;
                    if (c == L'\\') {
                        if (c == L'a')
                            c = L'\a';
                        else if (c == L'b')
                            c = L'\b';
                        else if (c == L'f')
                            c = L'\f';
                        else if (c == L'n')
                            c = L'\n';
                        else if (c == L'r')
                            c = L'\r';
                        else if (c == L't')
                            c = L'\t';
                        else if (c == L'v')
                            c = L'\v';
                        else if (c == L'\\')
                            c = L'\\';
                        else if (c == L'\'')
                            c = L'\'';
                        else if (c == L'\"')
                            c = L'\"';
                        else if (c == L'?')
                            c = L'\?';
                        else if (c == L'\n') {
                            c = *(++*s);
                            if (c == L'\"') {
                                key[len] = L'\0';
                                break;
                            } else if (c == L'\0') {
                                json_node_destroy(node);
                                node = NULL;
                                array_not_end = 0;
                                _json_error_message = L"Can not find End of Key";
                                break;
                            }
                        } else if (c == L'\"') {
                            key[len] = L'\0';
                            break;
                        } else if (c == L'\0') {
                            json_node_destroy(node);
                            node = NULL;
                            array_not_end = 0;
                            _json_error_message = L"Can not find End of Key";
                            break;
                        }
                    } else if (c == L'\"') {
                        key[len] = L'\0';
                        break;
                    } else if (c == L'\0') {
                        json_node_destroy(node);
                        node = NULL;
                        array_not_end = 0;
                        _json_error_message = L"Can not find End of Key";
                        break;
                    }

                    key[len] = c;
                    ++len;
                    ++*s;
                }
                ++*s;
                break;
            }
            ++*s;
        }

        if (array_not_end) {
            while (1) {
                if (**s == L'\0') {
                    json_node_destroy(node);
                    node = NULL;
                    array_not_end = 0;
                    _json_error_message = L"Can not find ':' symbol";
                    break;
                } else if (**s == L':') {
                    break;
                }
                ++*s;
            }
            if (array_not_end) {
                while (1) {
                    if (**s == L'\0') {
                        json_node_destroy(node);
                        node = NULL;
                        array_not_end = 0;
                        _json_error_message = L"Can not find Element in Object";
                        break;
                    } else if (**s == L'n') {
                        element = _json_parse_null(s);
                        break;
                    } else if (**s == L't' || **s == 'f') {
                        element = _json_parse_boolean(s);
                        break;
                    } else if ((L'0' <= **s && **s <= '9')
                            || **s == L'-'
                            || **s == L'+'
                            || **s == L'.'
                            || **s == L'e'
                            || **s == L'E') {
                        element = _json_parse_number(s);
                        break;
                    } else if (**s == L'\"') {
                        element = _json_parse_string(s);
                        break;
                    } else if (**s == L'[') {
                        element = _json_parse_array(s);
                        break;
                    } else if (**s == L'{') {
                        element = _json_parse_object(s);
                        break;
                    }
                    ++*s;
                }
                if (array_not_end) {
                    if (element) {
                        json_object_set(node, key, element);
                        while (1) {
                            if (**s == L',') {
                                break;
                            }
                            if (**s == L'}') {
                                array_not_end = 0;
                                break;
                            }
                            if (**s == L'\0') {
                                json_node_destroy(node);
                                node = NULL;
                                array_not_end = 0;
                                _json_error_message = L"Can not find End of Object or next part";
                                break;
                            }
                            ++*s;
                        }
                    } else {
                        swprintf(_json_error_path + wcslen(_json_error_path), _json_string_lengths, L" \"%Ls\" <", key);
                        json_node_destroy(node);
                        node = NULL;
                        array_not_end = 0;
                        break;
                    }
                }
            }
        }
    }
    ++*s;
    free(key);

    return node;
}

static void _json_print_null(json_node *node, wchar_t **s)
{
    node->type = node->type;
    (*s)[0] = L'n';
    (*s)[1] = L'u';
    (*s)[2] = L'l';
    (*s)[3] = L'l';
    *s += 4;
    **s = L'\0';
}

static void _json_print_boolean(json_node *node, wchar_t **s)
{
    if (node->t.boolean) {
        (*s)[0] = L't';
        (*s)[1] = L'r';
        (*s)[2] = L'u';
        (*s)[3] = L'e';
        *s += 4;
    } else {
        (*s)[0] = L'f';
        (*s)[1] = L'a';
        (*s)[2] = L'l';
        (*s)[3] = L's';
        (*s)[4] = L'e';
        *s += 5;
    }
    **s = L'\0';
}

static void _json_print_number(json_node *node, wchar_t **s)
{
    swprintf(*s, _json_string_lengths, L"%lf", node->t.number);
    *s += wcslen(*s);
}

static void _json_print_string(json_node *node, wchar_t **s)
{
    size_t len, i;
    wchar_t c;

    **s = L'\"';
    ++*s;
    len = wcslen(node->t.string);
    for (i = 0; i < len; i++) {
        c = node->t.string[i];
        if (c == L'\a') {
            (*s)[0] = '\\';
            (*s)[1] = 'a';
            ++*s;
        } else if (c == L'\b') {
            (*s)[0] = '\\';
            (*s)[1] = 'b';
            ++*s;
        } else if (c == L'\f') {
            (*s)[0] = '\\';
            (*s)[1] = 'f';
            ++*s;
        } else if (c == L'\n') {
            (*s)[0] = '\\';
            (*s)[1] = 'n';
            ++*s;
        } else if (c == L'\r') {
            (*s)[0] = '\\';
            (*s)[1] = 'r';
            ++*s;
        } else if (c == L'\t') {
            (*s)[0] = '\\';
            (*s)[1] = 't';
            ++*s;
        } else if (c == L'\v') {
            (*s)[0] = '\\';
            (*s)[1] = 'v';
            ++*s;
        } else if (c == L'\\') {
            (*s)[0] = '\\';
            (*s)[1] = '\\';
            ++*s;
        } else if (c == L'\'') {
            (*s)[0] = '\\';
            (*s)[1] = '\'';
            ++*s;
        } else if (c == L'\"') {
            (*s)[0] = '\\';
            (*s)[1] = '"';
            ++*s;
        } else if (c == L'\?') {
            (*s)[0] = '\\';
            (*s)[1] = '?';
            ++*s;
        } else {
            (*s)[0] = c;
        }
        ++*s;
    }
    **s = L'\"';
    ++*s;
    **s = L'\0';
}

static void _json_print_array(json_node *node, wchar_t **s)
{
    json_node *curr;

    **s = L'[';
    ++*s;
    for (curr = node->t.list->head; curr != NULL; curr = curr->next)
    {
        if (curr->type == JSON_TYPE_NULL)
            _json_print_null(curr, s);
        else if (curr->type == JSON_TYPE_BOOLEAN)
            _json_print_boolean(curr, s);
        else if (curr->type == JSON_TYPE_NUMBER)
            _json_print_number(curr, s);
        else if (curr->type == JSON_TYPE_STRING)
            _json_print_string(curr, s);
        else if (curr->type == JSON_TYPE_ARRAY)
            _json_print_array(curr, s);
        else if (curr->type == JSON_TYPE_OBJECT)
            _json_print_object(curr, s);
        if (curr->next != NULL) {
            **s = L',';
            ++*s;
        }
    }
    **s = L']';
    ++*s;
    **s = L'\0';
}

static void _json_print_object(json_node *node, wchar_t **s)
{
    json_node *curr;
    size_t len, i;
    wchar_t c;

    **s = L'{';
    ++*s;
    for (curr = node->t.list->head; curr != NULL; curr = curr->next)
    {
        **s = L'\"';
        ++*s;
        len = wcslen(curr->key);
        for (i = 0; i < len; i++) {
            c = curr->key[i];
            if (c == L'\a') {
                (*s)[0] = '\\';
                (*s)[1] = 'a';
                ++*s;
            } else if (c == L'\b') {
                (*s)[0] = '\\';
                (*s)[1] = 'b';
                ++*s;
            } else if (c == L'\f') {
                (*s)[0] = '\\';
                (*s)[1] = 'f';
                ++*s;
            } else if (c == L'\n') {
                (*s)[0] = '\\';
                (*s)[1] = 'n';
                ++*s;
            } else if (c == L'\r') {
                (*s)[0] = '\\';
                (*s)[1] = 'r';
                ++*s;
            } else if (c == L'\t') {
                (*s)[0] = '\\';
                (*s)[1] = 't';
                ++*s;
            } else if (c == L'\v') {
                (*s)[0] = '\\';
                (*s)[1] = 'v';
                ++*s;
            } else if (c == L'\\') {
                (*s)[0] = '\\';
                (*s)[1] = '\\';
                ++*s;
            } else if (c == L'\'') {
                (*s)[0] = '\\';
                (*s)[1] = '\'';
                ++*s;
            } else if (c == L'\"') {
                (*s)[0] = '\\';
                (*s)[1] = '"';
                ++*s;
            } else if (c == L'\?') {
                (*s)[0] = '\\';
                (*s)[1] = '?';
                ++*s;
            } else {
                (*s)[0] = c;
            }
            ++*s;
        }
        **s = L'\"';
        ++*s;
        **s = L':';
        ++*s;
        if (curr->type == JSON_TYPE_NULL)
            _json_print_null(curr, s);
        else if (curr->type == JSON_TYPE_BOOLEAN)
            _json_print_boolean(curr, s);
        else if (curr->type == JSON_TYPE_NUMBER)
            _json_print_number(curr, s);
        else if (curr->type == JSON_TYPE_STRING)
            _json_print_string(curr, s);
        else if (curr->type == JSON_TYPE_ARRAY)
            _json_print_array(curr, s);
        else if (curr->type == JSON_TYPE_OBJECT)
            _json_print_object(curr, s);
        if (curr->next != NULL) {
            **s = L',';
            ++*s;
        }
    }
    **s = L'}';
    ++*s;
    **s = L'\0';
}

void json_set_string_length(size_t length)
{
    _json_string_lengths = length;
}

const wchar_t *json_error_get_path(void)
{
    return _json_error_path;
}

const wchar_t *json_error_get_message(void)
{
    return _json_error_message;
}

void json_error_clear(void)
{
    _json_error_message = L"No erros";
    _json_error_path[0] = L'\0';
}

json_node *json_parse(const wchar_t *s)
{
    json_node *node;
    size_t pos;

    node = NULL;
    pos = 0;
    while (1) {
        if (*s == L'n') {
            node = _json_parse_null(&s);
            break;
        } else if (*s == L'f' || *s == L't') {
            node = _json_parse_boolean(&s);
            break;
        } else if ((L'0' <= *s && *s <= '9')
                || *s == L'-'
                || *s == L'+'
                || *s == L'.'
                || *s == L'e'
                || *s == L'E') {
            node = _json_parse_number(&s);
            break;
        } else if (*s == L'\"') {
            node = _json_parse_string(&s);
            break;
        } else if (*s == L'[') {
            node = _json_parse_array(&s);
            break;
        } else if (*s == L'{') {
            node = _json_parse_object(&s);
            break;
        } else if (*s == L'\0') {
            _json_error_message = L"Can not find JSON";
            break;
        }
        ++s;
    }
    if (node == NULL) {
        swprintf(_json_error_path + wcslen(_json_error_path), _json_string_lengths, L" ROOT");
    }

    return node;
}

void json_stringify(json_node *node, wchar_t *s)
{
    if (node->type == JSON_TYPE_NULL)
        _json_print_null(node, &s);
    else if (node->type == JSON_TYPE_BOOLEAN)
        _json_print_boolean(node, &s);
    else if (node->type == JSON_TYPE_NUMBER)
        _json_print_number(node, &s);
    else if (node->type == JSON_TYPE_STRING)
        _json_print_string(node, &s);
    else if (node->type == JSON_TYPE_ARRAY)
        _json_print_array(node, &s);
    else if (node->type == JSON_TYPE_OBJECT) {
        _json_print_object(node, &s);
    }
}

json_node *json_node_create(json_type type)
{
    json_node *node;
    
    node = (json_node *)malloc(sizeof(json_node));
    node->type = type;
    node->upper = NULL;
    node->key = NULL;
    if (node->type == JSON_TYPE_STRING) {
        node->t.string = (wchar_t *)malloc(sizeof(wchar_t) * _json_string_lengths);
    } else if (node->type == JSON_TYPE_ARRAY || node->type == JSON_TYPE_OBJECT) {
        node->t.list = (struct _json_list *)malloc(sizeof(struct _json_list));
        node->t.list->head = NULL;
        node->t.list->tail = NULL;
        node->t.list->length = 0;

    }

    return node;
}


void json_node_destroy(json_node *node)
{
    json_node *curr;
    
    if (node->type == JSON_TYPE_STRING) {
        free(node->t.string);
    } else if (node->type == JSON_TYPE_ARRAY || node->type == JSON_TYPE_OBJECT) {
            curr = node->t.list->head;
            if (curr != NULL) {
                while ((curr = curr->next) != NULL)
                    json_node_destroy(curr->prev);
                json_node_destroy(node->t.list->tail);
            }
            free(node->t.list);
    }

    if (node->key != NULL)
        free(node->key);
    free(node);
}

json_type json_node_type(json_node *node)
{
    return node->type;
}

json_node *json_node_upper(json_node *node)
{
    return node->upper;
}

size_t json_array_size(json_node *node)
{
    return node->t.list->length;
}

json_node *json_object_get(json_node *node, const wchar_t *key)
{
    json_node *curr;
    int found;

    found = 0;
    for (curr = node->t.list->head; curr != NULL; curr = curr->next) {
        if (wcscmp(curr->key, key) == 0) {
            found = 1;
            break;
        }
    }
    if (found)
        return curr;
    return NULL;
}

void json_array_push_back(json_node *dest_node, json_node *src_node)
{
    src_node->next = NULL;
    src_node->prev = dest_node->t.list->tail;
    dest_node->t.list->tail->next = src_node;
    dest_node->t.list->tail = src_node;
    src_node->upper = dest_node;
    ++dest_node->t.list->length;
}

void json_array_push_front(json_node *dest_node, json_node *src_node)
{
    src_node->prev = NULL;
    src_node->next = dest_node->t.list->head;
    dest_node->t.list->head->prev = src_node;
    dest_node->t.list->head = src_node;
    src_node->upper = dest_node;
    ++dest_node->t.list->length;
}

json_node *json_array_pop_back(json_node *node)
{
    json_node *ret;

    ret = NULL;
    if (node->t.list->tail != NULL) {
        ret = node->t.list->tail;
        node->t.list->tail = node->t.list->tail->prev;
        if (node->t.list->tail == NULL)
            node->t.list->head = NULL;
        else
            node->t.list->tail->next = NULL;
        --node->t.list->length;
    }

    return ret;
}

json_node *json_array_pop_front(json_node *node)
{
    json_node *ret;

    ret = NULL;
    if (node->t.list->head != NULL) {
        ret = node->t.list->head;
        node->t.list->head = node->t.list->head->next;
        if (node->t.list->head == NULL)
            node->t.list->tail = NULL;
        else
            node->t.list->head->prev = NULL;
        --node->t.list->length;
    }

    return ret;
}


json_node *json_object_set(json_node *dest_node, const wchar_t *key, json_node *src_node)
{
    json_node *curr;
    json_node *old;
    int found;

    if (src_node->key == NULL)
        src_node->key = (wchar_t *)malloc(sizeof(wchar_t) * _json_string_lengths);
    wcscpy(src_node->key, key);
    src_node->upper = dest_node;
    old = NULL;
    found = 0;
    for (curr = dest_node->t.list->head; curr != NULL; curr = curr->next) {
        if (wcscmp(curr->key, key) == 0) {
            found = 1;
            break;
        }
    }
    if (found) {
        old = curr;
        if (curr->next != NULL)
            curr->next->prev = src_node;
        if (curr->prev != NULL)
            curr->prev->next = src_node;
        src_node->prev = curr->prev;
        src_node->next = curr->next;
    } else {
        curr = dest_node->t.list->tail;
        dest_node->t.list->tail = src_node;
        src_node->next = NULL;
        if (curr == NULL) {
            dest_node->t.list->head = src_node;
            src_node->prev = NULL;
        } else {
            curr->next = src_node;
            src_node->prev = curr;
        }
        ++dest_node->t.list->length;
    }

    return old;
}

double json_number_get_value(json_node *node)
{
    return node->t.number;
}

const wchar_t *json_string_get_value(json_node *node)
{
    return node->t.string;
}

int json_boolean_get_value(json_node *node)
{
    return node->t.boolean;
}

void json_number_set_value(json_node *node, double value)
{
    node->t.number = value;
}

void json_string_set_value(json_node *node, const wchar_t *s)
{
    wcscpy(node->t.string, s);
}

void json_boolean_set_value(json_node *node, int value)
{
    node->t.boolean = value;
}
